#!/usr/bin/env ruby

require 'optparse'
require 'fileutils'

TAIL_BUF_LENGTH = 1 << 10

$option_parser
$options = {}
$options[:userprefix] = ""
$options[:usersuffix] = ""
$options[:debug] = false
$required_options = [:svn_url, :conversion_root, :config_file]

# Name of option => File verification method.
$path_options = { :conversion_root => :directory?, :config_file => :file? }

$p4_svn_convert = ""

$svndump_file = "svn.dump"
$svncache_dir
$config_hash
$changemap_lastrev

def dputs(*argv)
  puts argv if $options[:debug]
end

def usage
  puts $option_parser
  exit
end

def verify_required_options
  $required_options.each do |required_option|
    if $options[required_option].nil?
      puts "#{$PROGRAM_NAME}: --#{required_option.to_s} must be specified"
      usage
    end
  end
end

def verify_path_options
  $path_options.each_pair do |path_opt, verify_method|
    path = $options[path_opt]
    unless File.exists?(path)
        puts "#{$PROGRAM_NAME}: \"#{path}\" does not exist."
        usage
    end

    unless File.send(verify_method, path)
        path_type = verify_method.to_s().sub(/[?]/, '')
        puts "#{$PROGRAM_NAME}: \"#{path}\" must be a #{path_type}."
        usage
    end
  end
end

def verify_svn_connection
  `svn info #{$options[:svn_url]}`
  if $?.to_i != 0
    puts "#{$PROGRAM_NAME}: Unable to connect to svn repo at \"#{$options[:svn_url]}\".  Exiting."
    exit 1
  end
  $svncache_dir = "#{$options[:conversion_root]}/svn_cache"
end

def parse_config_file
  config_lines = File.open($options[:config_file]).map do |l|
    l.sub(/[#].*/, '').strip
  end

  config_lines.select! { |l| l =~ /[^=]+=[^=]+$/ }

  $config_hash = Hash[*config_lines.map { |l| l.split('=') }.flatten]
end

def parse_changemap_file

  return unless File.exists?($config_hash['com.p4convert.log.changeMap'])

  p4rev = nil
  svnrev = nil
  File.open($config_hash['com.p4convert.log.changeMap']) do |f|

    curpos = f.size
    while curpos > 0
      curpos -= TAIL_BUF_LENGTH
      if curpos < 0
        curpos = 0
      end
      f.seek(curpos, IO::SEEK_SET)
      lines = f.read.split("\n")
      if lines.count > 1 || curpos == 0
        p4rev, svnrev = lines[-1].split(", ") 
        break;
      end
    end
  end

  $changemap_lastrev = svnrev.nil? ? nil : svnrev.to_i
end

def update_config_file_key(key, value)
  `sed -i '' -e 's,^#{key}.*,#{key}=#{value},' #{$options[:config_file]}`
end

def verify_starting_rev
  if !$changemap_lastrev.nil?
    $options[:starting_rev] = $changemap_lastrev + 1
  elsif !$config_hash['com.p4convert.svn.start'].nil?
    $options[:starting_rev] = $config_hash['com.p4convert.svn.start'].to_i
  else 
    puts "Starting revision not found\nPlease verify your config file #{$options[:config_file]}"
    exit 1
  end

  update_config_file_key('com.p4convert.svn.start', $options[:starting_rev])
end

def verify_ending_rev
  if !$options[:max_revisions].nil? && !$options[:ending_rev].nil?
    puts "--max_revisons and --ending_revision are mutially exclusive."
    exit 1
  end

  if !$options[:max_revisions].nil?
    $options[:ending_rev] = $options[:starting_rev] +
                            $options[:max_revisions] - 1
  end

  if $options[:ending_rev] > 0 &&
     $options[:ending_rev] < $options[:starting_rev]
    puts "Ending revision must be greater than or equal to starting revision."
    exit 1
  end

  update_config_file_key('com.p4convert.svn.end', $options[:ending_rev])
end

def sync_svn_repo
  unless File.exists?($svncache_dir) && File.directory?($svncache_dir)
    dputs "Creating #{$svncache_dir}"
    `svnadmin create #{$svncache_dir}`
    if $?.to_i != 0
      puts "#{$PROGRAM_NAME}: Unable to create svn cache repo at \"#{$svncache_dir}\".  Exiting."
      FileUtils.rmtree($svncache_dir)
      exit 1
    end

    dputs "Installing Hooks"
    FileUtils.ln_s "/usr/bin/true", "#{File.join($svncache_dir, 'hooks', 'pre-revprop-change')}"

    dputs "Initializing #{$svncache_dir}"
    `svnsync --non-interactive --trust-server-cert initialize file://#{$svncache_dir} #{$options[:svn_url]}`
    if $?.to_i != 0
      puts "#{$PROGRAM_NAME}: Unable to initialize svn cache repo at \"#{$svncache_dir}\".  Exiting."
      FileUtils.rmtree($svncache_dir)
      exit 1
    end
  end

  dputs "Syncing #{$svncache_dir}"
  `svnsync --non-interactive --trust-server-cert sync file://#{$svncache_dir}`
  if $?.to_i != 0
    puts "#{$PROGRAM_NAME}: Unable to connect to svn sync from \"#{$options[:svn_url]}\".  Exiting."
    exit 1
  end

  svn_youngest_rev = `svnlook youngest #{$svncache_dir}`
  if $?.to_i != 0
    puts "#{$PROGRAM_NAME}: Unable to determine youngest revision in \"#{$svncache_dir}\".  Exiting."
    exit 1
  end

  if $options[:ending_rev] > svn_youngest_rev.to_i || $options[:ending_rev] == 0
    $options[:ending_rev] = svn_youngest_rev.to_i
    update_config_file_key('com.p4convert.svn.end', $options[:ending_rev])
  end
end

def dump_svn_revs
  dputs "Dumping #{$svncache_dir}"
  ending_rev = $options[:ending_rev] == 0 ? "HEAD" : $options[:ending_rev]
  `svnadmin dump #{$svncache_dir} --incremental -r #{$options[:starting_rev]}:#{ending_rev} > #{$svndump_file} 2>/dev/null`
  update_config_file_key('com.p4convert.svn.dumpFile', $svndump_file)
end

def build_users_map
  user_map = "#{$options[:conversion_root]}/users.map"
  File.unlink(user_map) if File.exists?(user_map)
  unless $options[:userprefix].empty? && $options[:usersuffix].empty?
    `#{$p4_svn_convert} --users #{$svndump_file} > #{user_map}`
    `sed -i '' -e 's/^\\([^#].*, \\)\\(.*\\)/\\1#{$options[:userprefix]}\\2#{$options[:usersuffix]}/' #{user_map}`
  end
end

def execute_import
  `#{$p4_svn_convert} #{$options[:config_file]}`
end

$option_parser = OptionParser.new do |o|
  o.banner = "Usage: #{$PROGRAM_NAME} [options]"

  o.on(:REQUIRED, "--conversion_root=ROOTPATH", String,
       "Path to p4svnconvert workspace.") do |rootpath|
    $options[:conversion_root] = rootpath
  end

  o.on(:REQUIRED, "--config_file=CONFIG", String,
       "Path to p4svnconvert workspace.") do |config|
    $options[:config_file] = config
  end

  o.on(:REQUIRED, "--svn_url=URL", String,
       "Location of source SVN repo.") do |url|
    $options[:svn_url] = url
  end

  o.on("--max_revisions=MAXREVS", Integer,
       "Limit of revisions to process") do |maxrevs|
    $options[:max_revisions] = maxrevs
  end

  o.on("--starting_revision=STARTREV", Integer,
       "The first SVN revision to import.") do |rev|
    $options[:starting_rev] = rev
  end

  o.on("--ending_revision=ENDREV", Integer,
       "The last SVN revision to import.") do |rev|
    $options[:ending_rev] = rev
  end

  o.on("--userprefix=PREFIX", String,
       "Preppend PREFIX to imported usernames") do |prefix|
    $options[:userprefix] = prefix
  end

  o.on("--usersuffix=SUFFIX", String,
       "Append SUFFIX to imported usernames") do |suffix|
    $options[:usersuffix] = suffix
  end

  o.on("--debug", "Enable diagnostic output") do |suffix|
    $options[:debug] = true
  end

  o.on_tail("-h", "--help", "Show this message") do
    puts o
    exit
  end

  o.parse!
end

class File
  # Perform an action while holding a file lock on +filename+.
  # Returns result of the operation if lock can be 
  def self.with_lock(filename)
    File.open(filename, File::RDWR|File::CREAT, 0644) do |lockf|
      unlocked = lockf.flock(File::LOCK_EX|File::LOCK_NB)
      if unlocked
        lockf.truncate(0)
        lockf.puts Process.pid
        lockf.flush
        begin
          yield
        ensure
          lockf.truncate(0)
          lockf.flush
          lockf.flock(File::LOCK_UN)
        end
      else
        dputs "#{$PROGRAM_NAME}: Unable to acquire lock.  Exiting."
        1
      end
    end
  end
end

# Verify as much as possible before starting long running operations.
verify_required_options
verify_path_options
verify_svn_connection

$p4_svn_convert = "/usr/local/openjdk7/bin/java -Dlog4j.configuration=file:#{$options[:conversion_root]}/p4convert-svn/debug.log4j.xml -jar #{$options[:conversion_root]}/p4convert-svn/p4convert-svn.jar"

File.with_lock("#{$options[:conversion_root]}/lockfile") do
  parse_config_file
  parse_changemap_file
  verify_starting_rev
  verify_ending_rev
  
  sync_svn_repo

  if $options[:ending_rev] >= $options[:starting_rev]
    dump_svn_revs
    build_users_map
    execute_import
  end
end
